/*
 * copap3.c: 
 * 
 * C-MEX file for solving COPAP3    
 *          
 * USAGE
 *
 * [P,minCost,sOpt,kOpt]=copap3(C,dumCost)
 *
 * INPUTS
 *
 * C = m x n cost matrix 
 * dumCost = cost of assigning a point to a dummy point
 * 
 * OUTPUTS
 *
 * P = 1 x n permutation vector: P(i) = row assigned to i'th column
 * minCost = cost of optimal assignment
 * sOpt = optimal shift
 * kOpt = size of matching
 *
 *
 * Created 6/22/2005 by Clay Scott
 * Memory management functions borrowed from Don Waagen
 * Last uptated 6/22/2005 by Clay Scott
 *
 */ 

#include "mex.h"
#include <stdio.h>
#include <stdlib.h>

#define HUGE 1000000000000    /* very large number */

/* 
 * Global variables.  They are global so that we don't have to 
 * allocate/deallocate space for them with every call of lopap.
 */

double **Cdub;
double **L;
int **A;
int m,n;
double dumCost;

/*
 * Memory allocation utilities
 */ 


/* ********************************************************************
// function: mxCalloc2d
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************/

double **
mxCalloc2d(int rows, int cols)
{
   double *tmp;
   double **tmp2;
   int i;

   tmp  = (double *)mxCalloc(rows * cols,sizeof(double));
   tmp2 = (double**)mxCalloc(rows,sizeof(double *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

// ********************************************************************
// function: mxCalloc2di
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers of ints ....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************

int **
mxCalloc2di(int rows, int cols)
{
   int *tmp;
   int **tmp2;
   int i;

   tmp  = (int *)mxCalloc(rows * cols,sizeof(int));
   tmp2 = (int **)mxCalloc(rows,sizeof(int *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

/* ********************************************************************
// function: mxFree2d
// 
// Purpose: Deallocate memory for 2 dimensional matrix (double or float)
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2d() or mxCallocf2d()!
// *******************************************************************/

void
mxFree2d(double **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/* ********************************************************************
// function: mxFree2di
// 
// Purpose: Deallocate memory for 2 dimensional matrix of integers
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2di() 
// *******************************************************************/

void
mxFree2di(int **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/*
 * Some handy functions for min and max
 */ 
 
double min(double x, double y)
{
  return(x <= y ? x : y);
}
double max(double x, double y)
{
  return(x > y ? x : y);
}
double min3(double x, double y, double z)
{
  return(min(x,min(y,z)));
}
double max3(double x, double y, double z)
{
  return(max(x,max(y,z)));
}
/* Return index along with min value */
void min3i(double x, double y, double z, double *minval, int *minind)
{
  if (x <= y)
  {
    if (x <= z)
    {
      *minval = x;
      *minind = 1;
    }
    else
    {
      *minval = z;
      *minind = 3;
    }
  }
  else if (y <= z)
  {
    *minval = y;
    *minind = 2;
  }
  else
  {
    *minval = z;
    *minind = 3;
  }
}
/* Return index along with max value */
void max3i(double x, double y, double z, double *maxval, int *maxind)
{
  if (x > y)
  {
    if (x > z)
    {
      *maxval = x;
      *maxind = 1;
    }
    else
    {
      *maxval = z;
      *maxind = 3;
    }
  }
  else if (y > z)
  {
    *maxval = y;
    *maxind = 2;
  }
  else
  {
    *maxval = z;
    *maxind = 3;
  }
}

/*
 * Solve order preserving assignment problem
 */
void lopap3(int *s,int *xLo,int *xHi,int *x,double *cost)
{

    int i,j;      
    int *loLim;
    int *hiLim;
    double con1, con2, con3;
    double *u;
    int *v;
    
    extern double **Cdub;
    extern double **L;
    extern int **A;
    extern int m,n;
    extern double dumCost;
    
    u = (double *) mxCalloc(1,sizeof(double));
    v = (int *) mxCalloc(1,sizeof(int));

    /* Set limits */
    
    loLim = (int *) mxCalloc(m+1,sizeof(int));
    hiLim = (int *) mxCalloc(m+1,sizeof(int));
    for (i=0; i<m+1; i++)
        loLim[i] = max(xLo[i],*s);
    
    hiLim[0]=xHi[0];
    for (i=1; i<m; i++)
    {
        if (xHi[i+1]>xHi[i])
            hiLim[i]=xHi[i+1]-1;
        else
            hiLim[i]=xHi[i];
    }
    hiLim[m]=xHi[m];
    for (i=0; i<m+1; i++)
        hiLim[i]=min(n+*s,hiLim[i]);

    
    /* LOPAP: main loop */
    
    for (i=1; i<m+1; i++)
    { 
      for (j=loLim[i]; j<=hiLim[i]; j++)
      {
          if (j > loLim[i])
            con1=L[i][j-1];
          else
            con1=HUGE;
          if (j-1 >= loLim[i-1] & j-1 <= hiLim[i-1])
            con2=L[i-1][j-1]+Cdub[i-1][j-1];
          else
            con2=HUGE;
          if (j <= hiLim[i-1])
            con3=L[i-1][j]+dumCost;
          else
            con3=HUGE;
         
          min3i(con1, con2, con3, u, v);
   
          A[i][j]=*v;  
          L[i][j]=*u;
          
      } 
    }
 
    /* *********************************************
     * Retrace optimal path
     * ********************************************* */
    
    i=m;
    j=hiLim[i];
    *cost = L[i][j];
    x[i]=j;
    while (i > 0)
    {
        
        switch (A[i][j])
        {
            case 1:
                j=j-1;
                break;
            case 2:
                i=i-1;
                j=j-1;
                break;
            case 3:
                i=i-1;
                break;
            default: 
                printf("Switch statement 2 error lopap3: *v = %d\n",*v);
                return;
        }
        x[i]=j;
    }
/*    
    printf("s = %d, ",*s);
    printf("cost = %f\n",*cost); 

    if (*s==4)
    {
        for (i=0; i<m+1; i++)
        {
          printf("%d %d\n",loLim[i],hiLim[i]); 
        }
        printf("\n");
        for (i=0; i<m+1; i++)
        {
          printf("%d %d\n",xLo[i],xHi[i]); 
        }
        printf("\n");
    }
*/ 
}

/* Bisection */

void copapBisect(int *sLo,int *xLo,int *sHi,int *xHi,int *xOpt,double *cost,int *sBest)
{
    int i;
    int *sMid;
    int *sLoBest;
    int *sHiBest;
    int *xMid;
    int *xLoNew;
    int *xHiNew;
    double *costLo;
    double *costMid;
    double *costHi;
    int *ind;
    
    extern int m,n;
    
    xMid = (int *) mxCalloc(m+1,sizeof(int));
    xLoNew = (int *) mxCalloc(m+1,sizeof(int));
    xHiNew = (int *) mxCalloc(m+1,sizeof(int));
    
    sMid = (int *) mxCalloc(1,sizeof(int));
    sLoBest = (int *) mxCalloc(1,sizeof(int));
    sHiBest = (int *) mxCalloc(1,sizeof(int));
    
    costLo = (double *) mxCalloc(1,sizeof(double));
    costMid = (double *) mxCalloc(1,sizeof(double));
    costHi = (double *) mxCalloc(1,sizeof(double));
    
    ind = (int *) mxCalloc(1,sizeof(int));
    
    if (*sLo == *sHi - 1)
    {
      *cost = HUGE;
      return;
    }
    *sMid = (*sLo + *sHi)/2;
    lopap3(sMid,xLo,xHi,xMid,costMid);
    copapBisect(sLo,xLo,sMid,xMid,xLoNew,costLo,sLoBest);
    copapBisect(sMid,xMid,sHi,xHi,xHiNew,costHi,sHiBest);

    min3i(*costLo,*costMid,*costHi,cost,ind);

          switch (*ind)
            {
            case 1:
                for (i=0; i<m+1; i++)
                    xOpt[i]=xLoNew[i];
                *sBest = *sLoBest;
                break;
            case 2:
                for (i=0; i<m+1; i++)
                    xOpt[i]=xMid[i];
                *sBest = *sMid;          
                break;
            case 3:
                for (i=0; i<m+1; i++)
                    xOpt[i]=xHiNew[i];
                *sBest = *sHiBest;
                break;
            default: 
                printf("Switch statement error in copapBisect: *ind = %d\n",*ind);
                return;
            }

}

/* ******************************************************************
 * Main routine
 * ****************************************************************** */

void copap3(double **C, int numRows, int numCols, double epsDum, double *P, double *minCost, double *sOpt, double *kOpt)
{
  int i,j,jnext,jmatch;
  int *xLo;
  int *xHi;
  int *xInit;
  int *xInitDub;
  int *xRest;
  int *xOpt;
  int *s0;
  int *sn;
  int *sRest;
  double *costInit;
  double *costRest;
  
  /* Global vars */
  extern double **Cdub;
  extern int m,n;
  extern double dumCost;
  extern double **L;
  extern int **A;
  
  m = numRows;
  n = numCols;
  dumCost = epsDum;
  
  Cdub = mxCalloc2d(m,2*n);
  for (i=0; i<m; i++)
    for (j=0; j<n; j++)
    {
      Cdub[i][j]=C[i][j];
      Cdub[i][j+n]=C[i][j];
    }  
  
  L = mxCalloc2d(m+1,2*n+1);
  for (j=0; j<2*n+1; j++)
    L[0][j]=0;

  A = mxCalloc2di(m+1,2*n+1);
  
  xLo = (int *) mxCalloc(m+1,sizeof(int));
  xHi = (int *) mxCalloc(m+1,sizeof(int));
  xOpt = (int *) mxCalloc(m+1,sizeof(int));
  xInit = (int *) mxCalloc(m+1,sizeof(int));
  xInitDub = (int *) mxCalloc(m+1,sizeof(int));
  xRest = (int *) mxCalloc(m+1,sizeof(int));
  costInit = (double *) mxCalloc(1,sizeof(double));
  costRest = (double *) mxCalloc(1,sizeof(double));
  s0 = (int *) mxCalloc(1,sizeof(int));
  sn = (int *) mxCalloc(1,sizeof(int));
  sRest = (int *) mxCalloc(1,sizeof(int));
  
  /* ***************************************************
   * Initialize recursion
   * *************************************************** */
  
  for (i=0; i<m+1; i++)
  {
    xLo[i]=0;
    xHi[i]=n;
  }
  
  *s0 = 0;
  lopap3(s0,xLo,xHi,xInit,costInit);
    
  /* ***************************************************
   * Recursive bisection
   * *************************************************** */
  
  for (i=0; i<m+1; i++)
      xInitDub[i]=xInit[i]+n;
  
  *sn = n;
  copapBisect(s0,xInit,sn,xInitDub,xRest,costRest,sRest);
  
  /* ***************************************************
   * Determine winner
   * *************************************************** */

  if (*costInit <= *costRest)
  {
    for (i=0; i<m+1; i++)
      xOpt[i] = xInit[i];
    *minCost = *costInit;
    *sOpt = 0;
  }
  else
  {
    for (i=0; i<m+1; i++)
      xOpt[i] = xRest[i];
    *minCost = *costRest;
    *sOpt = *sRest;      
  }
  
 
  
  /* ***************************************************
   * Convert to permutation and determine matching size
   * *************************************************** */
  
  *kOpt = 0;
  for (j=0; j<n; j++)
    P[j]=0;
  for (i=0; i<m; i++)
  {
    j=xOpt[i];
    jnext=xOpt[i+1];
    if (j < jnext)
    {
      jmatch = jnext-1;
      if (jmatch >= n)
        jmatch = jmatch-n;
      P[jmatch]=i+1;
      *kOpt = *kOpt + 1;
    }
  }
  
  /* ***************************************************
   * Complete matching by adding dummy vars
   * *************************************************** */
  
  i=m+1;
  for (j=0; j<n; j++)
      if (P[j]==0)
        {
          P[j]=i;
          i=i+1;
        }
  
  
  mxFree2d(Cdub);
  mxFree2d(L);
  mxFree2di(A);

  
}

/* ****************************************
 * Mex gateway function for copap
 * **************************************** */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
  double *origC;    /* Input: square matrix of scores (in vector form) */
  double **C;        /* Scores converted to matrix form */
  int m, n;         /* number of source/target points; dims of C */
  double dumCost;  /* Input: cost of assigning to dummy point */
  double *P;           /* Output: optimal assignment */
  double *minCost;        /* Output: cost of optimal assignment */
  int *sOpt;        /* Optimal shift */
  int *kOpt;         /* Optimal matching size */
  int mrows, ncols; /* For error checking */
  int i,j;
  
  /* Check number of arguments */

  if (nrhs != 2)  
    mexErrMsgTxt("Two input arguments required."); 
  if (nlhs != 4) 
    mexErrMsgTxt("Four output arguments required."); 
   
  /* Error checking for first argument */
 
  m = mxGetM(prhs[0]);      /* Number of rows */
  n = mxGetN(prhs[0]);      /* Number of cols */
  
  if (!mxIsNumeric(prhs[0])) 
    mexErrMsgTxt("Argument must be numeric.");
  if (mxIsComplex(prhs[0]))
    mexErrMsgTxt("Argument must be real.");
  if (mxGetNumberOfDimensions(prhs[0]) > 2) 
    mexErrMsgTxt("Input must be a two-dimensional matrix");  
  
  /* Error checking for second argument */
    
  mrows = mxGetM(prhs[1]);
  ncols = mxGetN(prhs[1]);
  if(!mxIsDouble(prhs[1]) || mxIsComplex(prhs[1]) || !(mrows == 1 && ncols == 1) ) 
    mexErrMsgTxt("Second input must be a noncomplex double");
      
  /* Assign (pointers to) inputs */
  
  origC = (double *) mxGetPr(prhs[0]);
  dumCost = (double) mxGetScalar(prhs[1]);
  
  /* Additional error checking */
  
  if (m <=0 || n <= 0)
    mexErrMsgTxt("C is degenerate.");
  if (dumCost <= 0)
    mexErrMsgTxt("Second argument should be positive ");
  
  /* Assign pointers to outputs */
  
  plhs[0] = mxCreateDoubleMatrix(1,n,mxREAL);
  plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
  plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
  plhs[3] = mxCreateDoubleMatrix(1,1,mxREAL);
  P = (double *) mxCalloc(n, sizeof(double));
  P = mxGetPr(plhs[0]);
  minCost = mxGetPr(plhs[1]);
  sOpt=(int *) mxGetPr(plhs[2]);
  kOpt=(int *) mxGetPr(plhs[3]);

  /* Convert costs from linear array to matrix form */
  
  C = mxCalloc2d(m,n);
  for (i=0; i<m; i++)
    for (j=0; j<n; j++) 
      C[i][j]=origC[i+m*j];
 
  /* Call main routine */
    
  copap3(C, m, n, dumCost, P, minCost, sOpt, kOpt);
  
} /* end of mexFunction */
