/*
 * copap3.c: 
 * 
 * Created 6/22/2005 by Clay Scott
 *
 */ 

#include "mex.h"
#include <stdio.h>
#include <stdlib.h>

#define EPS 0.000000001    /* very small number */


/*
 * Memory allocation utilities
 */ 


/* ********************************************************************
// function: mxCalloc2d
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************/

double **
mxCalloc2d(int rows, int cols)
{
   double *tmp;
   double **tmp2;
   int i;

   tmp  = (double *)mxCalloc(rows * cols,sizeof(double));
   tmp2 = (double**)mxCalloc(rows,sizeof(double *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

// ********************************************************************
// function: mxCalloc2di
// 
// Purpose: Allocate memory for 2 dimensional matrix.
//
// Returns: Pointer to array of pointers of ints ....
//
// Notes:  Underlying matrix is one dimensional, with pointers pointing
// to correct locations.  Makes mxFree()'ing easy.
// ********************************************************************

int **
mxCalloc2di(int rows, int cols)
{
   int *tmp;
   int **tmp2;
   int i;

   tmp  = (int *)mxCalloc(rows * cols,sizeof(int));
   tmp2 = (int **)mxCalloc(rows,sizeof(int *));
   
   for (i = 0; i < rows; i++)
      tmp2[i] = &tmp[i*cols];  /* pointer assignment */
   return(tmp2);
}

/* ********************************************************************
// function: mxFree2d
// 
// Purpose: Deallocate memory for 2 dimensional matrix (double or float)
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2d() or mxCallocf2d()!
// *******************************************************************/

void
mxFree2d(double **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/* ********************************************************************
// function: mxFree2di
// 
// Purpose: Deallocate memory for 2 dimensional matrix of integers
//
// Returns: void
//
// Notes:  Use *ONLY* with matrices created with mxCalloc2di() 
// *******************************************************************/

void
mxFree2di(int **matrix)
{
   mxFree(matrix[0]);
   mxFree(matrix);
}

/* ******************************************************************
 * Main routine
 * ****************************************************************** */

void get_costmat(double **BH1, double **BH2, int nsamp1, int nsamp2, int nbins, double **C)
{

  int i,j, k;
  double *BH1sum;
  double *BH2sum;
  
  /* normalize histograms */
  
  BH1sum = (double *) mxCalloc(nsamp1,sizeof(double));
  BH2sum = (double *) mxCalloc(nsamp2,sizeof(double));
  
  for (i=0; i<nsamp1; i++)
  {
    BH1sum[i]=0;  
    for (k=0; k<nbins; k++)
        BH1sum[i]=BH1sum[i]+BH1[i][k];
    for (k=0; k<nbins; k++)
        BH1[i][k]=BH1[i][k]/(BH1sum[i]+EPS);
  }
  
  for (j=0; j<nsamp2; j++)
  {
    BH2sum[j]=0;  
    for (k=0; k<nbins; k++)
        BH2sum[j]=BH2sum[j]+BH2[j][k];
    for (k=0; k<nbins; k++)
        BH2[j][k]=BH2[j][k]/(BH2sum[j]+EPS);
  }
  
  /* Compute chi-squared distances */
  
  for (i=0; i<nsamp1; i++)
  {  
    for (j=0; j<nsamp2; j++)
    {  
      C[i][j]=0;  
      for (k=0; k<nbins; k++)
        C[i][j]=C[i][j]+(BH1[i][k]-BH2[j][k])*(BH1[i][k]-BH2[j][k])/(BH1[i][k]+BH2[j][k]+EPS);
      C[i][j]=.5*C[i][j];
/*      printf("%f ",C[i][j]); */
    }
/*    printf("\n"); */
  }
}

/* ****************************************
 * Mex gateway function for copap
 * **************************************** */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
  double *mat1; /* Shape context one */
  double *mat2; /* Shape context two */
  double **BH1;    /* mat1 in matrix form */
  double **BH2;     /* mat2 in matrix form */
  double *Cout;     /* Returned matrix in linear array form */
  double **C;        /* Scores converted to matrix form */
  int nsamp1, nsamp2;         /* number of samples */
  int nbins1, nbins2;
  int i,j;
  
  /* Check number of arguments */

  if (nrhs != 2)  
    mexErrMsgTxt("Two input arguments required."); 
  if (nlhs != 1) 
    mexErrMsgTxt("Four output arguments required."); 
   
  /* Error checking for first argument */
 
  nsamp1 = mxGetM(prhs[0]);      /* Number of rows */
  nbins1 = mxGetN(prhs[0]);      /* Number of cols */
  
  if (!mxIsNumeric(prhs[0])) 
    mexErrMsgTxt("Argument must be numeric.");
  if (mxIsComplex(prhs[0]))
    mexErrMsgTxt("Argument must be real.");
  if (mxGetNumberOfDimensions(prhs[0]) > 2) 
    mexErrMsgTxt("Input must be a two-dimensional matrix");  
  
  /* Error checking for second argument */
    
  nsamp2 = mxGetM(prhs[1]);      /* Number of rows */
  nbins2 = mxGetN(prhs[1]);      /* Number of cols */
  
  if (!mxIsNumeric(prhs[1])) 
    mexErrMsgTxt("Argument must be numeric.");
  if (mxIsComplex(prhs[1]))
    mexErrMsgTxt("Argument must be real.");
  if (mxGetNumberOfDimensions(prhs[1]) > 2) 
    mexErrMsgTxt("Input must be a two-dimensional matrix");  
      
  /* Assign (pointers to) inputs */
  
  mat1 = (double *) mxGetPr(prhs[0]);
  mat2 = (double *) mxGetPr(prhs[1]);
  
  /* Additional error checking */
  
  if (nsamp1 <=0 || nsamp2 <= 0)
    mexErrMsgTxt("One of the SC matrices is degenerate.");
  if (nbins1 != nbins2)
    mexErrMsgTxt("Shape contexts have different sizes");
  

  /* Convert shape context data from linear array to matrix form */
  
  BH1 = mxCalloc2d(nsamp1,nbins1);
  for (i=0; i<nsamp1; i++)
    for (j=0; j<nbins1; j++) 
      BH1[i][j]=mat1[i+nsamp1*j];
 
  BH2 = mxCalloc2d(nsamp2,nbins2);
  for (i=0; i<nsamp2; i++)
    for (j=0; j<nbins2; j++) 
      BH2[i][j]=mat2[i+nsamp2*j];
  
  C = mxCalloc2d(nsamp1,nsamp2);

  /* Call main routine */
    
  get_costmat(BH1, BH2, nsamp1, nsamp2, nbins1, C);
  
  /* Assign pointers to outputs */
  
  plhs[0] = mxCreateDoubleMatrix(nsamp1,nsamp2,mxREAL);
  Cout = (double *) mxCalloc(nsamp1*nsamp2, sizeof(double));
  Cout = mxGetPr(plhs[0]);

  /* Convert cost matrix back to linear array */

  for (i=0; i<nsamp1; i++)
      for (j=0; j<nsamp2; j++)
          Cout[i+nsamp1*j]=C[i][j];
  
  
  mxFree2d(BH1);
  mxFree2d(BH2);
  mxFree2d(C);

} /* end of mexFunction */
