
function [BH1]=get_sc(V1tmp,nsamp)
%
% Get shape context for shape in image V1, based on nsamp samples
%
% USAGE
%
% [BH1]=get_sc(V1,nsamp)
%
% INPUTS
%
% V1: greyscale or binary images of objects to be compared.
% nsampe: number of samples along boundary of object
%
% OUTPUT
%
% BH1: vector representing a shape context log-polar histogram
%
% NOTES
%
% The original code on which this is based was taken from
% Serge Belongie's web site.
%
% 10/29/02 Clay Scott
%



%
% Parameters I occasionally tweak
%
tang_meth = 0;      % method for computing tangents: 0 -> from image gradient, 1 -> from boundary points (needs work)
mean_dist_global=[];    % for shape context scale normalization: 
                        % set equal to [] for scale invariance
convex = 0;         % 1 -> object of interest is known to be convex, 0 -> not

%
% Parameters I rarely tweak
%
nbins_theta=12;
nbins_r=5;
r_inner=1/8;
r_outer=2;

[N1,N2]=size(V1tmp);
V1=zeros(N1+2,N2+2);
V1(2:N1+1,2:N2+1)=V1tmp;
V1=V1/max(V1(:));

%
% Get boundary points
%
th = .5;
[x1,y1,t1]=bdry(V1,th,tang_meth);

% Note from Clay: bdry uses contourc to get the boundary
% pts.  The pts are in 'ij' coordinates, which are like row/column
% coordinates.  x1 and x2 are the column coords, and y1 and y2
% are the row coords.  However, the shape context/transformation 
% code treats these as if they are xy coordinates.  It doesn't 
% affect the outcome since we are consistent.
% Hence, when we display the sampled points
% below, these aren't really what we're working with.
% The only thing to be aware of is that the points being used for
% the computations are actually
% oriented in a counter-clockwise fashion, when viewed in xy coords,
% even though the tanget vectors point in a clockwise orientation.


% ensure clockwise orientation (MATLAB's contourc isn't consistent 
% in this regard).  Need this for COPAP matching.
ang_diff1=diff([t1; t1(1)]);
ang_ind1=find(abs(ang_diff1)>pi);
if sum(sign(ang_diff1(ang_ind1))) < 0
   x1=flipud(x1);
   y1=flipud(y1);
   t1=flipud(t1); 
end

% select subset of boundary points
nsamp1=length(x1);
if nsamp1==0
  error('No object in image')
  return;    
end
if nsamp1>=nsamp
   [x1,y1,t1]=get_samples(x1,y1,t1,nsamp);

else
    nsamp=nsamp1;
    % NOW ALLOWING FEWER THAN TARGET NUMBER
    %   error('shape #1 doesn''t have enough samples')
end
X=[x1 y1];

out_vec_1=zeros(1,nsamp);  % indicates outliers in shape 1

% compute shape context histograms
[BH1,mean_dist_1]=sc_compute(X',t1',mean_dist_global,nbins_theta,nbins_r,r_inner,r_outer,out_vec_1,convex);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% newfun: Extract boundary points
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [x,y,t,c]=bdry(V,th,tang_meth);
% [x,y,t,c]=bdry(V,th);
% extract a set of boundary points w/oriented tangents

if 1
    W=5;    % window width
    h=fspecial('gaussian',W);   % create filter
    Vp=padarray(V,[(W-1)/2,(W-1)/2],'symmetric','both');
    Vg=filter2(h,Vp,'valid');
else
    Vg=V; % if no smoothing is needed
end
c=contourc(Vg,[th th]);
[G1,G2]=gradient(Vg);

% if multiple contours, take largest.

cind=find(c(1,:)==th);
[cnum, kind]=max(c(2,cind));
bind=cind(kind);
B=c(:,bind+1:bind+cnum);
   
x=B(1,:)';
y=B(2,:)';

%plot(x,y)
%axis ij

% compute tangets
npts=size(B,2);
t=zeros(npts,1);
if tang_meth==0
 for n=1:npts
   x0=round(B(1,n)); % consider rounding coord of B to match tan vecs
   y0=round(B(2,n));
   %keyboard;
   t(n)=atan2(G2(y0,x0),G1(y0,x0))+pi/2;
 end
else
  yfor=[y; y(1)];
  xfor=[x; x(1)];
  yback=[y(end); y];
  xback=[x(end); x];
  yfor_diff=diff(yfor);
  xfor_diff=diff(xfor);
  yback_diff=diff(yback);
  xback_diff=diff(xback);
  theta_for=atan2(yfor_diff,xfor_diff);
  theta_back=atan2(yback_diff,xback_diff);
  for_dist=sqrt(xfor_diff.^2 + yfor_diff.^2);
  back_dist=sqrt(xback_diff.^2 + yback_diff.^2);
  for_frac=back_dist./(for_dist+back_dist);
  back_frac=for_dist./(for_dist+back_dist);
  t=mod(back_frac.*theta_back+for_frac.*theta_for+pi*(theta_for.*theta_back >= 0),2*pi);
 % if angles have opposite sign, can't just add them.  must also rotate pi radians.
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% newfun: Sample boundary points
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [xi,yi,ti,ind0]=get_samples(x,y,t,nsamp);
% [xi,yi,ti]=get_samples(x,y,t,nsamp);
%
% Sample boundary points with uniform spacing 

N=length(x);

ind0=1:N/nsamp:N+1;
ind0(end)=[];
ind0=round(ind0);
xi=x(ind0);
yi=y(ind0);
ti=t(ind0);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% newfun: Compute shape context
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [BH,mean_dist,r_array]=sc_compute(Bsamp,Tsamp,mean_dist,nbins_theta,nbins_r,r_inner,r_outer,out_vec,convex);
% [BH,mean_dist]=sc_compute(Bsamp,Tsamp,mean_dist,nbins_theta,nbins_r,r_inner,r_outer,out_vec);
%
% compute (r,theta) histograms for points along boundary 
%
% Bsamp is 2 x nsamp (x and y coords.)
% Tsamp is 1 x nsamp (tangent theta)
% out_vec is 1 x nsamp (0 for inlier, 1 for outlier)
%
% mean_dist is the mean distance, used for length normalization
% if it is not supplied, then it is computed from the data
%
% outliers are not counted in the histograms, but they do get
% assigned a histogram
%

nsamp=size(Bsamp,2);
in_vec=out_vec==0;

% compute r,theta arrays
r_array=real(sqrt(dist2(Bsamp',Bsamp'))); % real is needed to
                                          % prevent bug in Unix version
theta_array_abs=atan2(Bsamp(2,:)'*ones(1,nsamp)-ones(nsamp,1)*Bsamp(2,:),Bsamp(1,:)'*ones(1,nsamp)-ones(nsamp,1)*Bsamp(1,:))';
theta_array=theta_array_abs-Tsamp'*ones(1,nsamp); % orient with respect to tanget vector

% create joint (r,theta) histogram by binning r_array and
% theta_array

% normalize distance by mean, ignoring outliers
if isempty(mean_dist)
   tmp=r_array(in_vec,:);
   tmp=tmp(:,in_vec);
   mean_dist=mean(tmp(:));
end
r_array_n=r_array/mean_dist;

% use a log. scale for binning the distances
r_bin_edges=logspace(log10(r_inner),log10(r_outer),nbins_r);
r_array_q=zeros(nsamp,nsamp);
for m=1:nbins_r
    r_array_q=r_array_q+(r_array_n<r_bin_edges(m)); 
end
fz=r_array_q>0; % flag all points inside outer boundary

% put all angles in [0,2pi) range
theta_array_2 = rem(rem(theta_array,2*pi)+2*pi,2*pi);
% quantize to a fixed set of angles (bin edges lie on 0,(2*pi)/k,...2*pi
theta_array_q = 1+floor(theta_array_2/(2*pi/nbins_theta));

if convex
    fz = fz & ((theta_array_q <= floor(nbins_theta/2) - 1 ) | (theta_array_q == nbins_theta));
end

nbins=nbins_theta*nbins_r;
BH=zeros(nsamp,nbins);
for n=1:nsamp
   fzn=fz(n,:)&in_vec;
   Sn=sparse(theta_array_q(n,fzn),r_array_q(n,fzn),1,nbins_theta,nbins_r);
   BH(n,:)=Sn(:)';
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% newfun: Useful function for computing distances
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function n2 = dist2(x, c)
%DIST2	Calculates squared distance between two sets of points.
%
%	Description
%	D = DIST2(X, C) takes two matrices of vectors and calculates the
%	squared Euclidean distance between them.  Both matrices must be of
%	the same column dimension.  If X has M rows and N columns, and C has
%	L rows and N columns, then the result has M rows and L columns.  The
%	I, Jth entry is the  squared distance from the Ith row of X to the
%	Jth row of C.
%
%	See also
%	GMMACTIV, KMEANS, RBFFWD
%

%	Copyright (c) Christopher M Bishop, Ian T Nabney (1996, 1997)

[ndata, dimx] = size(x);
[ncentres, dimc] = size(c);
if dimx ~= dimc
	error('dist2.m: Data dimension does not match dimension of centres')
end

n2 = (ones(ncentres, 1) * sum((x.^2)', 1))' + ...
  		ones(ndata, 1) * sum((c.^2)',1) - ...
  		2.*(x*(c'));

